           (*-------------------------------------------*
            |                   Test                    |
            |                                           |
            |        CSP-Prover on Isabelle2009         |
            |                   June 2009               |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Test_model_T
imports CSP_T_Main
begin

(*****************************************************************

         1. simple example for model T
         2. 
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                         event
 *********************************************************)

datatype Event = a | b | c
datatype RunName = Run
datatype AName = A

(*********************************************************
            specification SPC and system IMP
 *********************************************************)

(*** Spc ***)

consts 
  Runfun :: "RunName => (RunName, Event) proc"

primrec
  "Runfun   Run = ? x -> $Run"

defs (overloaded)
Set_Runfun_def [simp]: "PNfun == Runfun"

(*** Imp ***)

consts 
  Afun :: "AName => (AName, Event) proc"

primrec
  "Afun      A = a -> $A"

defs (overloaded)
Set_Afun_def [simp]: "PNfun == Afun"

(*********************************************************
            relation between SPC and IMP
 *********************************************************)

consts 
  Run_to_A :: "RunName => (AName, Event) proc"

primrec
  "Run_to_A Run = $A"

(*********************************************************
               guardedfun (rutine work)
 *********************************************************)

(*** To automatically unfold syntactic sugar ***)

lemma guardedfun_Spc_Imp[simp]:
      "guardedfun Runfun"
      "guardedfun Afun"
by (simp add: guardedfun_def, rule allI, induct_tac p, simp_all)+

(*********************************************************
                   ? SPC <=F IMP ?
 *********************************************************)

defs FPmode_def [simp]: "FPmode == CMSmode"
(* it declares to use CMS approach.

   If you want to verify them by CPO approach, 
   use the following mode:

defs FPmode_def [simp]: "FPmode == CPOmode"

   In this example, both modes are available,
   because Spcfun and Impfun are guarded.       *)

lemma "($Run::(RunName, Event) proc) <=T $A"

  (***** by fixed point induction *****)
apply (rule cspT_fp_induct_left[of _ "Run_to_A"])

  (***** check guarded and no hiding operators *****)

apply (simp_all)
apply (simp)

  (***** by step laws (for transforming it to a hnf) *****)

apply (induct_tac p)
apply (cspT_auto)+
apply (cspT_choice)          (* replace ? by ! *)
apply (rule cspT_decompo_ref)
apply (auto)
done

end
